import import_string

from chamber.models import SmartModel
from chamber.utils.datastructures import ChoicesNumEnum
from django.core.exceptions import ValidationError
from django.db import models
from django.utils.translation import ugettext, ugettext_lazy as _
from django.template import Template, Context
from django.template.exceptions import TemplateSyntaxError, TemplateDoesNotExist

from pymess.config import settings, get_email_sender
from pymess.utils.html import raise_error_if_contains_banned_tags

from .common import BaseAbstractTemplate, BaseMessage, BaseRelatedObject


__all__ = (
    'EmailMessage',
    'EmailRelatedObject',
    'EmailTemplate',
    'Attachment',
    'AbstractEmailTemplate',
    'EmailTemplateDisallowedObject',
)


class EmailMessage(BaseMessage):

    STATE = ChoicesNumEnum(
        ('WAITING', _('waiting'), 1),
        ('SENDING', _('sending'), 2),
        ('SENT', _('sent'), 3),
        ('ERROR', _('error'), 4),
        ('DEBUG', _('debug'), 5),
    )

    recipient = models.EmailField(verbose_name=_('recipient'), blank=False, null=False)
    template = models.ForeignKey(settings.EMAIL_TEMPLATE_MODEL, verbose_name=_('template'), blank=True, null=True,
                                 on_delete=models.SET_NULL, related_name='email_messages')
    state = models.IntegerField(verbose_name=_('state'), null=False, blank=False, choices=STATE.choices, editable=False)
    sender = models.EmailField(verbose_name=_('sender'), blank=False, null=False)
    sender_name = models.CharField(verbose_name=_('sender name'), blank=True, null=True, max_length=250)
    subject = models.TextField(verbose_name=_('subject'), blank=False, null=False)
    number_of_send_attempts = models.PositiveIntegerField(verbose_name=_('number of send attempts'), null=False,
                                                          blank=False, default=0)
    last_webhook_received_at = models.DateTimeField(
        verbose_name=_('last webhook received at'),
        null=True,
        blank=True,
        editable=False,
    )
    info_changed_at = models.DateTimeField(
        verbose_name=_('info changed at'),
        null=True,
        blank=True,
        editable=False,
    )

    @property
    def friendly_sender(self):
        """
        returns sender with sender name in standard address format if sender name was defined
        """
        return '{} <{}>'.format(self.sender_name, self.sender) if self.sender_name else self.sender

    @property
    def failed(self):
        return self.state == self.STATE.ERROR

    def __str__(self):
        return '{}: {}'.format(self.recipient, self.subject)

    class Meta(BaseMessage.Meta):
        verbose_name = _('e-mail message')
        verbose_name_plural = _('e-mail messages')


class EmailRelatedObject(BaseRelatedObject):

    email_message = models.ForeignKey(EmailMessage, verbose_name=_('e-mail message'), null=False, blank=False,
                                      on_delete=models.CASCADE, related_name='related_objects')

    class Meta(BaseRelatedObject.Meta):
        verbose_name = _('related object of a e-mail message')
        verbose_name_plural = _('related objects of e-mail messages')


class AttachmentManager(models.Manager):

    def create_from_tripple(self, tripple):
        filename, file, content_type = tripple
        attachment = self.model(
            email_message=self.instance,
            content_type=content_type
        )
        attachment.file.save(filename, file, save=True)
        return attachment

    def create_from_tripples(self, *tripples):
        return [
            self.create_from_tripple(tripple) for tripple in tripples
        ]


class Attachment(SmartModel):

    email_message = models.ForeignKey(EmailMessage, verbose_name=_('e-mail message'), on_delete=models.CASCADE,
                                      related_name='attachments')
    content_type = models.CharField(verbose_name=_('content type'), blank=False, null=False, max_length=100)
    file = models.FileField(verbose_name=_('file'), null=False, blank=False, upload_to='pymess/emails')

    objects = AttachmentManager()

    def __str__(self):
        return '#{}'.format(self.pk)

    class Meta:
        verbose_name = _('attachment')
        verbose_name_plural = _('attachments')


class AbstractEmailTemplate(BaseAbstractTemplate):

    subject = models.TextField(verbose_name=_('subject'), blank=False, null=False)
    sender = models.EmailField(verbose_name=_('sender'), null=True, blank=True, max_length=200)
    sender_name = models.CharField(verbose_name=_('sender name'), blank=True, null=True, max_length=250)

    def _update_context_data(self, context_data):
        for context_processor_fun_name in settings.EMAIL_TEMPLATE_CONTEXT_PROCESSORS:
            context_data.update(import_string(context_processor_fun_name)(context_data, self))
        return context_data

    def clean_subject(self, context_data=None):
        try:
            self.render_subject(context_data or {})
        except (TemplateSyntaxError, TemplateDoesNotExist) as ex:
            raise ValidationError(ugettext('Error during template subject rendering: "{}"').format(ex))

    def get_subject(self):
        return self.subject

    def render_subject(self, context_data):
        context_data = self._update_context_data(context_data)
        return Template(self.get_subject()).render(Context(context_data))

    def get_backend_sender(self):
        return get_email_sender()

    def send(self, recipient, context_data, related_objects=None, tag=None, attachments=None, **kwargs):
        return super().send(
            recipient=recipient,
            context_data=context_data,
            related_objects=related_objects,
            tag=tag,
            sender=self.sender,
            subject=self.render_subject(context_data),
            sender_name=self.sender_name,
            attachments=attachments
        )

    class Meta(BaseAbstractTemplate.Meta):
        abstract = True
        verbose_name = _('e-mail template')
        verbose_name_plural = _('e-mail templates')


class EmailTemplate(AbstractEmailTemplate):

    def clean_body(self, context_data=None):
        super().clean_body(context_data={'EMAIL_DISABLE_VARIABLE_VALIDATOR': True})
        raise_error_if_contains_banned_tags(self.body)

    def clean_subject(self, context_data=None):
        super().clean_subject(context_data={'EMAIL_DISABLE_VARIABLE_VALIDATOR': True})
        raise_error_if_contains_banned_tags(self.subject)

    def _extend_body(self, template_body):
        base_template = settings.EMAIL_TEMPLATE_BASE_TEMPLATE
        templatetags = settings.EMAIL_TEMPLATE_TEMPLATETAGS

        template_content = '{{% block {} %}}{}{{% endblock %}}'.format(
            settings.EMAIL_TEMPLATE_CONTENT_BLOCK,
            template_body
        )

        out = []
        if base_template is not None:
            out.append('{{% extends \'{}\' %}}'.format(base_template))
        if templatetags:
            out.append('{{% load {} %}}'.format(' '.join(templatetags)))
        out.append(template_content)

        return ''.join(out)

    def get_body(self):
        return self._extend_body(self.body) if settings.EMAIL_TEMPLATE_EXTEND_BODY else self.body


class EmailTemplateDisallowedObject(BaseRelatedObject):

    template = models.ForeignKey(
        verbose_name=_('template'),
        to=EmailTemplate,
        null=True, blank=True,
        on_delete=models.CASCADE,
        related_name='disallowed_objects',
        db_index=True
    )

    class Meta(BaseRelatedObject.Meta):
        verbose_name = _('disallowed object of an e-mail template')
        verbose_name_plural = _('disallowed objects of e-mail templates')



import logging

from datetime import timedelta

from django.db import DatabaseError
from django.db.models import Q
from django.core.management.base import BaseCommand, CommandError
from django.utils.timezone import now

from pymess.config import get_email_sender, settings
from pymess.models import EmailMessage

from chamber.utils.transaction import atomic_with_signals


logger = logging.getLogger(__name__)


class Command(BaseCommand):
    """
    Command for sending e-mails in batch.
    """

    @atomic_with_signals
    def handle(self, *args, **options):
        if not settings.EMAIL_BATCH_SENDING:
            raise CommandError('Batch sending is turned off')

        email_sender = get_email_sender()

        sent = 0
        try:
            messages_to_send = EmailMessage.objects.filter(
                Q(state=EmailMessage.STATE.WAITING) |
                Q(
                    state=EmailMessage.STATE.ERROR,
                    number_of_send_attempts__lte=settings.EMAIL_BATCH_MAX_NUMBER_OF_SEND_ATTEMPTS,
                    created_at__gte=now() - timedelta(seconds=settings.EMAIL_BATCH_MAX_SECONDS_TO_SEND)
                )
            ).select_for_update(nowait=True).order_by('created_at')[:settings.EMAIL_BATCH_SIZE]
            for message in messages_to_send:
                try:
                    email_sender.publish_message(message)
                    sent += 1
                except Exception as ex:
                    # Rollback shoud not be applied for already send e-mails
                    logger.exception(ex)
            self.stdout.write('{} e-mails sent'.format(sent))
        except DatabaseError:
            self.stdout.write('e-mail messages are already locked')




from chamber.exceptions import PersistenceException
from django.core.exceptions import ValidationError
from django.utils import timezone as tz
from django.utils.encoding import force_text
from django.utils.translation import ugettext as _

from pymess.backend.dialer import DialerBackend, DialerSendingError
from pymess.config import settings
from pymess.models import DialerMessage
from pymess.utils.logged_requests import generate_session


class DaktelaDialerBackend(DialerBackend):
    """
    Dialer backend implementing Daktela service https://www.daktela.com/api/v6/models/campaignsrecords
    """

    SESSION_SLUG = 'pymess-daktela_autodialer'

    @staticmethod
    def _get_dialer_api_url(name=None):
        name = '/{name}'.format(name=name) if name else ''

        return '{base_url}{name}.json?accessToken={access_token}'.format(
            base_url=settings.DIALER_DAKTELA.URL, name=name, access_token=settings.DIALER_DAKTELA.ACCESS_TOKEN,
        )

    def _get_autodialer_message_states(self, resp_json, state_mapped, resp_message_state):
        custom_fields = resp_json['result']['customFields']
        tts_processed = custom_fields['ttsprocessed'][0]
        whole_message_heard = ('whole_message_heard' in custom_fields and custom_fields['whole_message_heard']
                               and custom_fields['whole_message_heard'][0] == 'Yes')
        if state_mapped == DialerMessage.STATE.ANSWERED_PARTIAL and whole_message_heard:
            resp_message_state = str(DialerMessage.STATE.ANSWERED_COMPLETE)
        if state_mapped == DialerMessage.STATE.DONE and tts_processed == '0':
            resp_message_state = str(DialerMessage.STATE.NOT_ASSIGNED)
        tts_processed = resp_json['result']['customFields']['ttsprocessed'][0]
        is_final_state = resp_json['result']['action'] == '5' and tts_processed == '1'
        message_state = settings.DIALER_DAKTELA.STATES_MAPPING[resp_message_state]
        return is_final_state, message_state

    def _update_dialer_states(self, messages):
        """
        Method uses Daktela API to get info about autodialer call status
        :param messages: list of dialer messages to update
        """
        for message in messages:
            name = message.extra_data['name']
            client_url = self._get_dialer_api_url(name)
            response = generate_session(
                slug=self.SESSION_SLUG,
                related_objects=(message,),
                timeout=settings.DIALER_DAKTELA.TIMEOUT
            ).get(client_url)
            resp_json = response.json()

            message.extra_data.update({
                'name': name,
                'daktela_action': resp_json['result']['action'],
                'daktela_statuses': resp_json['result']['statuses'],
            })
            resp_message_state = resp_json['result']['statuses'][0]['name'] if len(
                resp_json['result']['statuses']) else resp_json['result']['action']
            state_mapped = settings.DIALER_DAKTELA.STATES_MAPPING[resp_message_state]
            message_error = resp_json['error'] if len(resp_json['error']) else None

            if message.is_autodialer:
                is_final_state, message_state = self._get_autodialer_message_states(resp_json, state_mapped,
                                                                                    resp_message_state)
            else:
                message_state = state_mapped
                is_final_state = resp_json['result']['action'] == '5'

            try:
                self.update_message(
                    message,
                    state=message_state,
                    error=message_error,
                    extra_data=message.extra_data,
                    is_final_state=is_final_state,
                )
            except Exception as ex:
                self.update_message(message, state=DialerMessage.STATE.ERROR, error=force_text(ex), is_final_state=True)
                # Do not re-raise caught exception. We do not know exact exception to catch so we catch them all
                # and log them into database. Re-raise exception causes transaction rollback (lost of information about
                # exception).

    def _update_message_with_error(self, message, error_message):
        self.update_message(
            message,
            state=DialerMessage.STATE.ERROR,
            error=error_message,
            is_final_state=True
        )

    def publish_message(self, message):
        """
        Method uses Daktela API for sending dialer message
        :param message: dialer message
        """
        client_url = self._get_dialer_api_url()
        testing_phones = settings.DIALER_DAKTELA.TESTING_PHONES
        if testing_phones and message.recipient not in testing_phones:
            error_message = _('Not allowed number to dial. Allowed numbers are: {}').format(', '.join(testing_phones))
            self._update_message_with_error(message, error_message)
            return
        try:
            payload = {
                'queue': (settings.DIALER_DAKTELA.AUTODIALER_QUEUE if message.is_autodialer
                          else settings.DIALER_DAKTELA.PREDICTIVE_QUEUE),
                'number': message.recipient,
                'customFields': {
                    'mall_pay_text': [
                        message.content,
                    ],
                    'ttsprocessed': [
                        0,
                    ],
                },
                'action': 5,
            }
            custom_fields = message.extra_data.get('custom_fields')
            if custom_fields:
                payload['customFields'].update(**custom_fields)

            response = generate_session(
                slug=self.SESSION_SLUG,
                related_objects=(message,),
                timeout=settings.DIALER_DAKTELA.TIMEOUT
            ).post(
                client_url,
                json=payload,
            )
            resp_json = response.json()

            message.extra_data.update({
                'name': resp_json['result']['name'],
                'daktela_action': resp_json['result']['action'],
                'daktela_statuses': resp_json['result']['statuses'],
            })

            self.update_message(
                message,
                state=DialerMessage.STATE.READY,
                error=resp_json['error'] if len(resp_json['error']) else None,
                sent_at=tz.now(),
                extra_data=message.extra_data,
            )
        except Exception as ex:
            self._update_message_with_error(message, force_text(ex))
            # Do not re-raise caught exception. We do not know exact exception to catch so we catch them all
            # and log them into database. Re-raise exception causes transaction rollback (lost of information about
            # exception).

    def create_message(self, recipient, content=None, related_objects=None, tag=None, template=None,
                       is_autodialer=True, **kwargs):
        """
        Create dialer message which will be logged in the database (content is not needed for this .
        :param recipient: phone number of the recipient
        :param related_objects: list of related objects that will be linked with the dialer message using generic
        relation
        :param tag: string mark that will be saved with the message
        :param kwargs: extra attributes that will be saved with the message
        """
        try:
            return super().create_message(
                recipient=recipient,
                content=content,
                related_objects=related_objects,
                tag=tag,
                template=template,
                state=self.get_initial_dialer_state(recipient),
                is_autodialer=is_autodialer,
                **kwargs,
                **self._get_extra_message_kwargs()
            )
        except PersistenceException as ex:
            raise DialerSendingError(force_text(ex))



from functools import reduce

from operator import or_ as OR

from chamber.models import SmartModel
from chamber.utils.datastructures import ChoicesNumEnum
from django.conf import settings
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ValidationError
from django.db import models
from django.db.models import Manager, Q
from django.template import Template, Context
from django.template.exceptions import TemplateSyntaxError, TemplateDoesNotExist
from django.utils.translation import ugettext_lazy as _

from jsonfield.fields import JSONField

from pymess.utils import has_int_pk


class RelatedObjectManager(Manager):

    def create_from_related_object(self, related_object):
        return self.create(
            object_id_int=related_object.pk if has_int_pk(related_object) else None,
            object_id=related_object.pk,
            content_type=ContentType.objects.get_for_model(related_object)
        )

    def create_from_related_objects(self, *related_objects):
        return [
            self.create_from_related_object(related_object)
            for related_object in related_objects
        ]

    def _get_related_object_pks(self, *related_objects):
        return [related_object.pk for related_object in related_objects]

    def _get_related_objects_qs_kwargs(self, *related_objects):
        return reduce(OR, (
            Q(**{
                'object_id': obj.pk,
                'content_type': ContentType.objects.get_for_model(obj)
            }) for obj in related_objects
        ))

    def update_related_objects(self, *related_objects, limit_to_model=None):
        delete_qs = self.all()
        if limit_to_model is not None:
            delete_qs = delete_qs.filter(content_type=ContentType.objects.get_for_model(limit_to_model))
        if related_objects:
            delete_qs = delete_qs.exclude(self._get_related_objects_qs_kwargs(*related_objects))
        delete_qs.delete()
        return self.get_or_create_from_related_objects(*related_objects)

    def get_or_create_from_related_object(self, related_object):
        return self.get_or_create(
            object_id_int=related_object.pk if has_int_pk(related_object) else None,
            object_id=related_object.pk,
            content_type=ContentType.objects.get_for_model(related_object)
        )

    def get_or_create_from_related_objects(self, *related_objects):
        return [
            self.get_or_create_from_related_object(related_object)
            for related_object in related_objects
        ]

    def get_related_objects_by_model(self, model):
        values_list_pk = 'object_id_int' if has_int_pk(model) else 'object_id'
        pks = self.filter(content_type=ContentType.objects.get_for_model(model)).values_list(values_list_pk)
        return model.objects.filter(pk__in=pks)

    def filter_from_related_objects(self, *related_objects):
        if hasattr(self, 'instance'):
            template_filter = (Q(template=self.instance) | Q(template__isnull=True))
        else:
            template_filter = Q(template__isnull=True)

        return self.model.objects.filter(template_filter).filter(self._get_related_objects_qs_kwargs(*related_objects))


class BaseMessage(SmartModel):

    STATE = ChoicesNumEnum()

    sent_at = models.DateTimeField(verbose_name=_('sent at'), null=True, blank=True, editable=False)
    recipient = models.CharField(verbose_name=_('recipient'), null=False, blank=False, max_length=20)
    content = models.TextField(verbose_name=_('content'), null=False, blank=False)
    template_slug = models.SlugField(verbose_name=_('slug'), max_length=100, null=True, blank=True, editable=False)
    state = models.IntegerField(verbose_name=_('state'), null=False, blank=False, choices=STATE.choices, editable=False)
    backend = models.CharField(verbose_name=_('backend'), null=True, blank=True, editable=False, max_length=250)
    error = models.TextField(verbose_name=_('error'), null=True, blank=True, editable=False)
    extra_data = JSONField(verbose_name=_('extra data'), null=True, blank=True, editable=False)
    extra_sender_data = JSONField(verbose_name=_('extra sender data'), null=True, blank=True, editable=False)
    tag = models.SlugField(verbose_name=_('tag'), null=True, blank=True, editable=False)

    def __str__(self):
        return self.recipient

    class Meta:
        abstract = True
        ordering = ('-created_at',)


class BaseRelatedObject(SmartModel):

    content_type = models.ForeignKey(ContentType, verbose_name=_('content type of the related object'),
                                     null=False, blank=False, on_delete=models.CASCADE)
    object_id = models.TextField(verbose_name=_('ID of the related object'), null=False, blank=False)
    object_id_int = models.PositiveIntegerField(verbose_name=_('ID of the related object in int format'), null=True,
                                                blank=True, db_index=True)
    content_object = GenericForeignKey('content_type', 'object_id')

    objects = RelatedObjectManager()

    class Meta:
        abstract = True
        ordering = ('-created_at',)

    def __str__(self):
        return str(self.content_object)


class BaseAbstractTemplate(SmartModel):

    slug = models.SlugField(verbose_name=_('slug'), max_length=100, null=False, blank=False, editable=False,
                            primary_key=True)
    body = models.TextField(verbose_name=_('message body'), null=True, blank=False)
    is_active = models.BooleanField(null=False, blank=False, default=True, verbose_name=_('is active'))
    is_allowed_duplicate_messages = models.BooleanField(null=False, blank=False, default=True,
                                                        verbose_name=_('Duplicate messages are allowed'))

    def _update_context_data(self, context_data):
        return context_data

    def render_text_template(self, text, context_data):
        context_data = self._update_context_data(context_data)
        return Template(text).render(Context(context_data))

    def render_body(self, context_data):
        return self.render_text_template(self.get_body(), context_data)

    def clean_body(self, context_data=None):
        try:
            self.render_body(context_data or {})
        except (TemplateSyntaxError, TemplateDoesNotExist) as ex:
            raise ValidationError(_('Error during template body rendering: "{}"').format(ex))

    def get_body(self):
        return self.body

    def can_send_for_object(self, related_objects):
        return (
            (
                self.is_allowed_duplicate_messages
                or not self.exist_duplicate_messages(related_objects)
            )
            and (
                not hasattr(self, 'disallowed_objects')
                or related_objects is None
                or not self.disallowed_objects.filter_from_related_objects(*related_objects).exists()
            )
        )

    def exist_duplicate_messages(self, related_objects):
        if related_objects:
            qs = self.get_backend_sender().model.objects.filter(template=self)
            for obj in related_objects:
                qs = qs.filter(
                    related_objects__object_id=obj.pk,
                    related_objects__content_type=ContentType.objects.get_for_model(obj)
                )
            return qs.exists()
        else:
            return False

    def can_send(self, recipient, related_objects):
        return self.is_active and self.can_send_for_object(related_objects)

    def get_backend_sender(self):
        raise NotImplementedError

    def send(self, recipient, context_data, related_objects=None, tag=None, **kwargs):
        if self.can_send(recipient, related_objects):
            return self.get_backend_sender().send(
                recipient=recipient,
                content=self.render_body(context_data),
                related_objects=related_objects,
                tag=tag,
                template=self,
                **kwargs,
            )
        else:
            return None

    def __str__(self):
        return self.slug

    class Meta:
        abstract = True
        ordering = ('-created_at',)



from pymess.backend.push import PushNotificationBackend
from pymess.models import PushNotificationMessage


class DummyPushNotificationBackend(PushNotificationBackend):
    """Dummy push notification backend used for testing environments. Backend only logs messages to the database."""

    def publish_message(self, message):
        self.update_message(message, state=PushNotificationMessage.STATE.DEBUG)
